@@ -59,6 +59,7 @@
 #include <ctime>
 #include <fstream>
 #include <queue>
+#include <limits>
 #include "LSDStatsTools.hpp"
 #include "LSDChiNetwork.hpp"
 #include "LSDRaster.hpp"
@@ -94,6 +95,33 @@
 #include <math.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <tuple>
+
+// struct link exampl;
+// {
+//   int source_node;
+//   int receiving_node;
+//   float zpass;
+//   int source_basin;
+//   int receiving_basin;
+// };
+// bool operator>( const link& lhs, const link& rhs )
+// {
+//   return lhs.zpass > rhs.zpass;
+// };
+// bool operator<( const link& lhs, const link& rhs )
+// {
+//   return lhs.zpass < rhs.zpass;
+// };
+// bool operator>=( const link& lhs, const link& rhs )
+// {
+//   return lhs.zpass >= rhs.zpass;
+// };
+// bool operator<=( const link& lhs, const link& rhs )
+// {
+//   return lhs.zpass <= rhs.zpass;
+// };
+
 
 // This empty constructor is just there to have a default one.
 void LSDDEM_xtensor::create()
@@ -150,6 +178,11 @@ void LSDDEM_xtensor::create(int tnrows, int tncols, float txmin, float tymin, fl
   JunctionNetwork = return_fake_junctionnetwork();
   ChiTool = return_fake_chitool();
 
+    // I just need that before
+  boundary_conditions = {"n","n","n","n"};
+  default_boundary_condition = true; // This is the default and recommended boundary conditions.
+  // Can be later explicitely modify
+
   // Object constructed
 }
 
@@ -181,9 +214,6 @@ void LSDDEM_xtensor::PreProcessing(bool carve, bool fill, float min_slope_for_fi
 // Calculates the FlowInfo object with D8 method
 void LSDDEM_xtensor::calculate_FlowInfo()
 {
-  // I just need that before
-  vector<string> temp_bc(4);
-  std::vector<std::string> boundary_conditions = {"n","n","n","n"};
   // Alright that's all I need, let's get the attributes now
   //# FlowInfo contains node ordering and all you need about navigatinfg through donor/receiver through your raster
   FlowInfo = LSDFlowInfo(boundary_conditions, PPRaster);
@@ -201,7 +231,7 @@ void LSDDEM_xtensor::calculate_FlowInfo_Dinf()
   std::cout << "IMPORTANT-WARNING: I havent tested the Dinf approach yet, carful with that." << std::endl;
   // I just need that before
   vector<string> temp_bc(4);
-  std::vector<std::string> boundary_conditions = {"n","n","n","n"};
+  boundary_conditions = {"n","n","n","n"};
   // Alright that's all I need, let's get the attributes now
   std::vector<LSDRaster> DINFRAST = PPRaster.D_inf_flowacc_DA();
   //# FlowInfo contains node ordering and all you need about navigatinfg through donor/receiver through your raster
@@ -323,6 +353,12 @@ void LSDDEM_xtensor::calculate_outlet_location_of_main_basin(bool test_drainage_
   JunctionNetwork.get_overlapping_channels_to_downstream_outlets(FlowInfo, baselevel_junctions, DistanceFromOutlet, sources_clean, outlet_nodes,baselevel_nodes,10);
 }
 
+void LSDDEM_xtensor::force_all_outlets(bool test_drainage_boundary)
+{
+  baselevel_junctions = JunctionNetwork.get_BaseLevelJunctions();
+  if(test_drainage_boundary)
+    baselevel_junctions = JunctionNetwork.Prune_Junctions_Largest(baselevel_junctions, FlowInfo, FlowAcc);
+}
 
 void LSDDEM_xtensor::generate_chi(float theta, float A_0)
 {
@@ -330,7 +366,7 @@ void LSDDEM_xtensor::generate_chi(float theta, float A_0)
   chi_coordinates = FlowInfo.get_upslope_chi_from_multiple_starting_nodes(baselevel_nodes, theta,A_0, min_cont);
   ChiTool.chi_map_automator_chi_only(FlowInfo, sources_clean, outlet_nodes, baselevel_nodes, PPRaster, DistanceFromOutlet, DrainageArea, chi_coordinates);
   // giving the basin informations to the object
-  vector<int> this_ordered_BL = ChiTool.get_ordered_baselevel();
+  std::vector<int> this_ordered_BL = ChiTool.get_ordered_baselevel();
   for(size_t i=0;i<this_ordered_BL.size();i++)
   {  
     baselevel_nodes_to_BK[this_ordered_BL[i]] = int(i);
@@ -651,6 +687,21 @@ xt::pytensor<float,2> LSDDEM_xtensor::get_hillshade(float hs_altitude, float hs_
   return output;
 }
 
+// Version of hillshading for custom raster
+xt::pytensor<float,2> LSDDEM_xtensor::get_hillshade_custom(xt::pytensor<float,2>& datrast, float hs_altitude, float hs_azimuth, float hs_z_factor)
+{
+
+  LSDRaster hs_raster, tohillshade = LSDRaster(nrows, ncols, xmin, ymin, cellsize, NoDataValue, xt_to_TNT(datrast,nrows,ncols));
+  hs_raster = tohillshade.hillshade(hs_altitude,hs_azimuth,hs_z_factor);
+  std::array<size_t,2> shape = {nrows,ncols};
+  xt::xtensor<int,2> output(shape);
+  for(size_t i=0;i<nrows;i++)
+  for(size_t j=0;j<ncols;j++)
+  {output(i,j) = hs_raster.get_data_element(i,j);}
+
+  return output;
+}
+
 
 std::vector<xt::pytensor<float,2> > LSDDEM_xtensor::get_polyfit_on_topo(float window, std::vector<int> selecao)
 {
@@ -754,9 +805,14 @@ void LSDDEM_xtensor::calculate_movern_disorder(float start_movern, float delta_m
   ChiTool_disorder.chi_map_automator_chi_only(FlowInfo, sources_clean, outlet_nodes, baselevel_nodes,
                         PPRaster, DistanceFromOutlet, DrainageArea, chi_coordinates);
 
+  // this map will store the uncertainties results
+  std::map<int, std::vector<float> > best_fit_movern_for_basins;
+  std::map<int, std::vector<float> > lowest_disorder_for_basins;
+
   //Getting the ordered base level
   std::vector<int> this_ordered_BL = ChiTool_disorder.get_ordered_baselevel();
 
+
   // creating the chi disorder vector of values
   std::vector<float> movern;
   for(int i=0; i<n_movern; i++)
@@ -778,10 +834,44 @@ void LSDDEM_xtensor::calculate_movern_disorder(float start_movern, float delta_m
     // recalculating chi for this movern
     LSDRaster this_chi = FlowInfo.get_upslope_chi_from_all_baselevel_nodes(movern[i], A_0, area_threshold);
     ChiTool_disorder.update_chi_data_map(FlowInfo, this_chi); // updating the chitool object
+
     // Potential multithreading her
     for(size_t BK =0; BK< this_ordered_BL.size(); BK++)
     {
       movern_per_BK[BK].push_back(ChiTool_disorder.test_collinearity_by_basin_disorder(FlowInfo, BK));
+      std::vector<float> disorder_stats = ChiTool_disorder.test_collinearity_by_basin_disorder_with_uncert(FlowInfo, BK);
+      // if this is the first m over n value, then initiate the vectors for this basin key
+      if (i == 0)
+      {
+        lowest_disorder_for_basins[BK] = disorder_stats;
+        
+        int n_combos_this_basin = int(disorder_stats.size());
+        std::vector<float> best_fit_movern;
+        for(int bf = 0; bf < n_combos_this_basin; bf++)
+        {
+          best_fit_movern.push_back( movern[i] );
+        }
+        best_fit_movern_for_basins[BK] = best_fit_movern;
+      }
+      else
+      {
+        // loop through all the combos and get the best fit movern
+        std::vector<float> existing_lowest_disorder = lowest_disorder_for_basins[BK];
+        std::vector<float> existing_best_fit_movern = best_fit_movern_for_basins[BK];
+        int n_combos_this_basin = int(disorder_stats.size());
+        
+        for(int bf = 0; bf < n_combos_this_basin; bf++)
+        {
+          if (existing_lowest_disorder[bf] > disorder_stats[bf] )
+          {
+            existing_lowest_disorder[bf] = disorder_stats[bf];
+            existing_best_fit_movern[bf] = movern[i];
+          }
+        }
+        lowest_disorder_for_basins[BK] = existing_lowest_disorder;
+        best_fit_movern_for_basins[BK] = existing_best_fit_movern;
+      }
+
     }
 
     if(verbose==true)
@@ -792,6 +882,8 @@ void LSDDEM_xtensor::calculate_movern_disorder(float start_movern, float delta_m
   {
     disorder_movern_per_BK[BK] = movern_per_BK[BK];
   }
+  best_fits_movern_per_BK = best_fit_movern_for_basins;
+  
   // done
 
 
@@ -809,9 +901,240 @@ xt::pytensor<float,1> LSDDEM_xtensor::burn_rast_val_to_xy(xt::pytensor<float,1>
   for(size_t i=0; i< X_coord.size(); i++)
   {
     PPRaster.get_row_and_col_of_a_point(X_coord[i], Y_coord[i], row, col);
-    output[i] = PPRaster.get_data_element(row,col);
+    float val = 0;
+    if(row == NoDataValue || col == NoDataValue)
+      val = NoDataValue;
+    else
+      val = PPRaster.get_data_element(row,col);
+    output[i] = val;
   }
   return output;
 }
 
+// Test function to return the perimeter of each basin, the "ridgeline"
+std::map<int, std::map<std::string, xt::pytensor<float,1> > > LSDDEM_xtensor::extract_perimeter_of_basins()
+{
+  std::map<int, std::map<std::string, xt::pytensor<float,1> > > output;
+
+  size_t n_bas = baselevel_junctions.size();
+  for(size_t samp = 0; samp<n_bas; samp++)
+  {
+    
+    std::cout << "Analysing catchment ..." << std::endl;
+    LSDBasin thisBasin(baselevel_junctions[samp],FlowInfo, JunctionNetwork);
+    std::cout << "Getting the perimeter ..." << std::endl;
+    thisBasin.set_Perimeter(FlowInfo);
+    std::vector<int> permineter_nodes = thisBasin.get_Perimeter_nodes();
+    std::cout << "Got it, let me format the data" << std::endl;
+    size_t n_node_in_this_perimeter = permineter_nodes.size();
+    std::array<size_t,1> sadfkhkjogjtaoehgoaisdfgoih = {n_node_in_this_perimeter};
+    xt::xtensor<float,1> X(sadfkhkjogjtaoehgoaisdfgoih), Y(sadfkhkjogjtaoehgoaisdfgoih), Z(sadfkhkjogjtaoehgoaisdfgoih);
+    int this_row, this_col; float this_X, this_Y, this_elev;
+    for(size_t huile_de_coude = 0; huile_de_coude < n_node_in_this_perimeter; huile_de_coude ++)
+    {
+      int this_node = permineter_nodes[huile_de_coude];
+      FlowInfo.retrieve_current_row_and_col(this_node,this_row,this_col);
+      FlowInfo.get_x_and_y_from_current_node(this_node,this_X,this_Y);
+      X[huile_de_coude] = this_X;Y[huile_de_coude] = this_Y;Z[huile_de_coude]= PPRaster.get_data_element(this_row,this_col); 
+    }
+
+    std::map<std::string, xt::pytensor<float,1> > temp_ou;
+    temp_ou["X"] = X;
+    temp_ou["Y"] = Y;
+    temp_ou["Z"] = Z;
+    output[BLjunctions_to_BK[baselevel_junctions[samp]]] = temp_ou;
+  }
+
+  return output;
+}
+
+
+std::map<std::string, xt::pytensor<float,1> > LSDDEM_xtensor::get_sources_full()
+{
+  std::array<size_t,1> sizla = { sources.size() };
+  xt::xtensor<float,1> outX(sizla), outY(sizla), outNodes(sizla);
+  for(size_t i=0;i<sources.size();i++)
+  {
+    int this_node = sources[i];
+    if(this_node <0)
+    {  
+          outX[i] = -9999;
+          outY[i]=- 9999;
+          outNodes[i] = -9999 ;
+          continue;
+    }
+    outNodes[i] = float(this_node);
+    float this_X,this_Y;
+    FlowInfo.get_x_and_y_from_current_node(this_node,this_X,this_Y);
+    outX[i] = this_X;
+    outY[i]=this_Y;
+  }
+  std::map<std::string, xt::pytensor<float,1> > output;
+  output["X"] = outX;
+  output["Y"] = outY;
+  output["nodeID"] = outNodes;
+  return output;
+}
+
+
+void LSDDEM_xtensor::calculate_discharge_from_precipitation(int tnrows, int tncols, float txmin, float tymin, float tcellsize, float tndv, xt::pytensor<float,2>& precdata, bool accumulate_current_node)
+{
+
+  // First step is to convert back the precipitation raster into LSDRaster
+  std::cout << "I am ingesting the precipitation raster. I am drinking it I guess? Right? haha " << std::endl;
+  LSDRaster precipitation_raster_original = LSDRaster(tnrows, tncols, txmin, tymin, tcellsize, tndv, xt_to_TNT(precdata, tnrows, tncols));
+  std::cout << "Got it" << std::endl;
+
+  // Potential recasting if extents are different
+  if(precipitation_raster_original.get_NRows()!= nrows || precipitation_raster_original.get_NCols()!= ncols)
+  {
+    std::cout << "I now need to recast it to the right extent. I am just resampling vertically, if you need a more fancy method just do it yourself and ingest a raster to the right extent yo!" << std::endl;
+    float goulg = 0;
+    TNT::Array2D<float> GORIS(nrows,ncols,goulg);
+    LSDRaster new_precipitation = LSDRaster(nrows, ncols, xmin, ymin, cellsize, NoDataValue, GORIS);; // copying the original raster
+    for(size_t i=0; i<new_precipitation.get_NRows(); i++)
+    for(size_t j=0; j<new_precipitation.get_NCols(); j++)
+    {
+      float x,y;
+      new_precipitation.get_x_and_y_locations(i,j,x,y);
+      float new_val = precipitation_raster_original.get_value_of_point(x,y);
+      new_precipitation.set_data_element(i,j,new_val);
+    }
+    std::cout << "RECASTED!" << std::endl;
+    precipitation_raster_original = new_precipitation;
+  }
+
+  std::cout << "Alright let's calculate the Discharge, it will take time" << std::endl;
+  float dx = precipitation_raster_original.get_DataResolution();
+
+  // volume precipitation per time precipitation times the cell areas
+  precipitation_raster_original.raster_multiplier(dx*dx);
+
+  // discharge accumulates this precipitation
+  DrainageArea = FlowInfo.upslope_variable_accumulator_v2(precipitation_raster_original, accumulate_current_node);
+  std::cout << "Done with that calculation! Note that I replace the drainage area raster!" << std::endl;
+
+}
+
+// Simple wrapper that returns xy coordinates from row and column
+std::pair<xt::pytensor<float,1>,xt::pytensor<float,1> >LSDDEM_xtensor::query_xy_from_rowcol(xt::pytensor<int,1>& row, xt::pytensor<int,1>& col)
+{
+  // first simply initialise the output
+  std::array<size_t,1> sizla = {row.size()};
+  xt::xtensor<float,1> Xs(sizla), Ys(sizla);
+
+  // then query my friend
+  for(size_t i=0; i< row.size(); i++)
+  {
+    float tx,ty;
+    PPRaster.get_x_and_y_locations(row[i],col[i],tx,ty);
+    Xs[i] = tx; Ys[i] = ty;
+  }
+
+  std::pair<xt::pytensor<float,1>,xt::pytensor<float,1> > output =  std::make_pair(Xs,Ys);  
+  return output;
+}
+
+
+std::map<std::string, xt::pytensor<float, 1> > LSDDEM_xtensor::get_FO_Mchi()
+{
+  std::map<std::string,  std::vector<int>  > out =  ChiTool.get_integer_vecdata_for_m_chi(FlowInfo);
+  std::map<std::string,  xt::pytensor<float, 1>  > recasted_output;
+  xt::pytensor<float, 1>  temp_vec_X,temp_vec_Y,temp_vec_FOMC;
+  // first I am creating the arrays
+  std::array<size_t,1> sizla = {out["nodeID"].size()};
+  xt::xtensor<float,1> this_X(sizla),this_Y(sizla),this_FOMC(sizla), this_elev(sizla), this_DA(sizla), this_flow(sizla);
+  for(size_t j = 0; j< out["nodeID"].size(); j++)
+  {
+    int this_node = out["nodeID"][j], row,col,recrow,reccol,recnode; FlowInfo.retrieve_current_row_and_col(this_node,row,col); FlowInfo.retrieve_receiver_information(this_node, recnode,recrow,reccol);
+    float dz,dchi,tx,ty, TFOCM;
+    if(this_node != recnode)
+    {
+      dchi = chi_coordinates.get_data_element(row,col) - chi_coordinates.get_data_element(recrow,reccol);
+      dz = PPRaster.get_data_element(row,col) - PPRaster.get_data_element(recrow,reccol);
+      this_FOMC[j] = dz/dchi;
+    }
+    else
+    {
+      this_FOMC[j] = 0;
+    }        
+    FlowInfo.get_x_and_y_locations(row,col,tx,ty);
+    this_X[j] = tx;this_Y[j]=ty;
+    this_elev[j] = PPRaster.get_data_element(row,col);
+    this_DA[j] = DrainageArea.get_data_element(row,col);
+    this_flow[j] = DistanceFromOutlet.get_data_element(row,col);
+    temp_vec_X = this_X;
+    temp_vec_Y = this_Y;
+    temp_vec_FOMC = this_FOMC;
+  }
+
+
+
+  recasted_output["X"] = temp_vec_X;
+  recasted_output["Y"] = temp_vec_Y;
+  recasted_output["fo_m_chi"] = temp_vec_FOMC;
+  recasted_output["DA"] = this_DA;
+  recasted_output["elevation"] = this_elev;
+  recasted_output["flow_distance"] = this_flow;
+
+  return recasted_output;
+}
+
+void LSDDEM_xtensor::mask_topo(float value, xt::pytensor<float,2> masker)
+{
+
+  for(size_t i = 0; i<nrows; i++)
+  {
+    for(size_t j = 0; j<ncols; j++)
+    {
+      if(masker(i,j) == value)
+        PPRaster.set_data_element(i,j,value);
+    }
+  }
+
+}
+
+
+std::tuple<std::vector<xt::pytensor<float,2> >, std::vector<std::map<std::string, float> > > LSDDEM_xtensor::get_individual_basin_raster()
+{
+  // formatting the outputs
+  std::vector<xt::pytensor<float,2> > vec_1;
+  std::vector<std::map<std::string, float> > vec_2;
+  // getting the base levels
+  for(size_t BK =0; BK< baselevel_junctions.size(); BK++)
+  {
+    std::cout << "DEBUG::PROC BAS " << BK << " FOR EXTRACTION" << std::endl; 
+    // First getting the basin
+    LSDBasin that_basin(baselevel_junctions[BK], FlowInfo,  JunctionNetwork);
+    // getting the array of topo values
+    LSDRaster trimmed_rast = that_basin.TrimPaddedRasterToBasin(30, FlowInfo,
+                                             PPRaster);
+    // Trimming the no data values
+    // LSDRaster trimmed_rast = this_rast.RasterTrimmerPadded(30);
+    // pushing in the output
+    vec_1.push_back(this->LSDRaster_to_xt(trimmed_rast));
+    // the saving method needs the following attribute: Z,x_min,x_max,y_min,y_max,res
+    std::vector<float> datvec =  trimmed_rast.get_XY_MinMax();
+    vec_2.push_back({{"x_min", datvec[0]}, {"y_min", datvec[1]}, {"x_max", datvec[2]}, {"y_max", datvec[3]}, {"res", trimmed_rast.get_DataResolution()} });
+  }
+
+  return {vec_1,vec_2};
+}
+
+xt::pytensor<float,2> LSDDEM_xtensor::LSDRaster_to_xt(LSDRaster& datrast)
+{
+
+  xt::pytensor<float,2> output = xt::zeros<float>({datrast.get_NRows(),datrast.get_NCols()});
+
+  for(size_t i=0; i< datrast.get_NRows(); i++)
+  {
+    for(size_t j=0; j<datrast.get_NCols() ; j++)
+    {
+      output(i,j) = datrast.get_data_element(i,j);
+    }
+  }
+  return output;
+}
+
+
 #endif
